==> dynamiq-0.3.2.tar.gz: dynamiq-0.3.2/dynamiq/components/evaluators/llm_evaluator.py <==
import json
from typing import Any
from warnings import warn

from dynamiq.nodes import Node
from dynamiq.prompts import Message, Prompt


class LLMEvaluator:
    """
    Uses an LLM to evaluate inputs based on a prompt containing instructions and examples.

    The inputs are lists that are user-defined depending on the desired metric. The output
    is a dictionary with a key `results` containing a list of evaluation results. Each result
    is a dictionary with user-defined keys and values of either 0 for FALSE or 1 for TRUE.

    Usage example:
    ```python
    from typing import List
    from dynamiq.components.evaluators import LLMEvaluator
    evaluator = LLMEvaluator(
        instructions="Is this answer problematic for children?",
        inputs=[("predicted_answers", List[str])],
        outputs=["score"],
        examples=[
            {"inputs": {"predicted_answers": "Damn, this is straight outta hell!!!"},
             "outputs": {"score": 1}},
            {"inputs": {"predicted_answers": "Football is the most popular sport."},
             "outputs": {"score": 0}},
        ],
    )
    predicted_answers = [
        "Football is the most popular sport with around 4 billion followers worldwide",
        "Python language was created by Guido van Rossum.",
    ]
    results = evaluator.run(predicted_answers=predicted_answers)
    print(results)
    # {'results': [{'score': 0}, {'score': 0}]}
    ```
    """

    def __init__(
        self,
        instructions: str,
        inputs: list[tuple[str, type[list]]],
        outputs: list[str],
        examples: list[dict[str, Any]],
        *,
        raise_on_failure: bool = True,
        llm: Node,
    ):
        """
        Creates an instance of LLMEvaluator.

        Args:
            instructions (str): The prompt instructions to use for evaluation.
            inputs (list[tuple[str, type[list]]]): The inputs that the component expects as incoming
                connections and that it evaluates. Each input is a tuple of an input name and input
                type. Input types must be lists.
            outputs (list[str]): Output names of the evaluation results. They correspond to keys in
                the output dictionary.
            examples (list[dict[str, Any]]): Few-shot examples conforming to the expected input and
                output format as defined in the `inputs` and `outputs` parameters. Each example is a
                dictionary with keys "inputs" and "outputs". They contain the input and output as
                dictionaries respectively.
            raise_on_failure (bool): If True, the component will raise an exception on an
                unsuccessful API call.
            llm (Node): The LLM node to use for evaluation.
        """
        self.validate_init_parameters(inputs, outputs, examples)
        self.raise_on_failure = raise_on_failure
        self.instructions = instructions
        self.inputs = inputs
        self.outputs = outputs
        self.examples = examples
        self.api_params = {}

        default_generation_kwargs = {
            "response_format": {"type": "json_object"},
            "seed": 42,
        }
        user_generation_kwargs = self.api_params.get("generation_kwargs", {})
        merged_generation_kwargs = {
            **default_generation_kwargs,
            **user_generation_kwargs,
        }
        self.api_params["generation_kwargs"] = merged_generation_kwargs

        template = self.prepare_template()
        message = Message(role="user", content=template)
        self.prompt = Prompt(messages=[message])
        self.builder = Prompt(messages=[message])

        self.llm = llm

    @staticmethod
    def validate_init_parameters(
        inputs: list[tuple[str, type[list]]],
        outputs: list[str],
        examples: list[dict[str, Any]],
    ):
        """
        Validate the init parameters.

        Args:
            inputs (list[tuple[str, type[list]]]): The inputs to validate.
            outputs (list[str]): The outputs to validate.
            examples (list[dict[str, Any]]): The examples to validate.

        Raises:
            ValueError: If the inputs are not a list of tuples with a string and a type of list.
                        If the outputs are not a list of strings.
                        If the examples are not a list of dictionaries.
                        If any example does not have keys "inputs" and "outputs" with values that
                        are dictionaries with string keys.
        """
        if (
            not isinstance(inputs, list)
            or not all(isinstance(_input, tuple) for _input in inputs)
            or not all(
                isinstance(_input[0], str)
                and _input[1] is not list
                and len(_input) == 2
                for _input in inputs
            )
        ):
            msg = (
                f"LLM evaluator expects inputs to be a list of tuples. Each tuple must contain "
                f"an input name and type of list but received {inputs}."
            )
            raise ValueError(msg)

        if not isinstance(outputs, list) or not all(
            isinstance(output, str) for output in outputs
        ):
            msg = f"LLM evaluator expects outputs to be a list of str but received {outputs}."
            raise ValueError(msg)

        if not isinstance(examples, list) or not all(
            isinstance(example, dict) for example in examples
        ):
            msg = f"LLM evaluator expects examples to be a list of dictionaries but received {examples}."
            raise ValueError(msg)

        for example in examples:
            if (
                {"inputs", "outputs"} != example.keys()
                or not all(
                    isinstance(example[param], dict) for param in ["inputs", "outputs"]
                )
                or not all(
                    isinstance(key, str)
                    for param in ["inputs", "outputs"]
                    for key in example[param]
                )
            ):
                msg = (
                    f"LLM evaluator expects each example to have keys `inputs` and `outputs` with "
                    f"values that are dictionaries with str keys but received {example}."
                )
                raise ValueError(msg)

    def run(self, **inputs) -> dict[str, Any]:
        """
        Run the LLM evaluator.

        Args:
            inputs: The input values to evaluate. The keys are the input names and the values are
                lists of input values.

        Returns:
            dict[str, Any]: A dictionary with a `results` entry that contains a list of results.
                Each result is a dictionary containing the keys as defined in the `outputs` parameter
                of the LLMEvaluator and the evaluation results as the values. If an exception occurs
                for a particular input value, the result will be `None` for that entry. If the API is
                "openai" and the response contains a "meta" key, the metadata from OpenAI will be
                included in the output dictionary, under the key "meta".

        Raises:
            ValueError: Only in the case that `raise_on_failure` is set to True and the received
                inputs are not lists or have different lengths, or if the output is not a valid JSON
                or doesn't contain the expected keys.
        """
        self.validate_input_parameters(dict(self.inputs), inputs)

        input_names, values = inputs.keys(), list(zip(*inputs.values()))
        list_of_input_names_to_values = [dict(zip(input_names, v)) for v in values]

        results: list[dict[str, Any] | None] = []
        errors = 0
        for input_names_to_values in list_of_input_names_to_values:
            prompt = self.prompt
            try:
                result = self.llm.execute(
                    input_data=input_names_to_values, prompt=prompt
                )
            except Exception as e:
                msg = (
                    f"Error while generating response for prompt: {prompt}. Error: {e}"
                )
                if self.raise_on_failure:
                    raise ValueError(msg)
                warn(msg)
                results.append(None)
                errors += 1
                continue

            if self.is_valid_json_and_has_expected_keys(
                expected=self.outputs, received=result["content"]
            ):
                parsed_result = json.loads(result["content"])
                results.append(parsed_result)
            else:
                results.append(None)
                errors += 1

        if errors > 0:
            msg = f"LLM evaluator failed for {errors} out of {len(list_of_input_names_to_values)} inputs."
            warn(msg)

        return {"results": results}

    def prepare_template(self) -> str:
        """
        Prepare the prompt template.

        Combine instructions, inputs, outputs, and examples into one prompt template with the
        following format:
        Instructions:
        <instructions>

        Generate the response in JSON format with the following keys:
        <list of output keys>
        Consider the instructions and the examples below to determine those values.

        Examples:
        <examples>

        Inputs:
        <inputs>
        Outputs:

        Returns:
            str: The prompt template.
        """
        inputs_section = (
            "{"
            + ", ".join(
                [
                    f'"{input_socket[0]}": {{{{ {input_socket[0]} }}}}'
                    for input_socket in self.inputs
                ]
            )
            + "}"
        )

        examples_section = "\n".join(
            [
                "Inputs:\n"
                + json.dumps(example["inputs"])
                + "\nOutputs:\n"
                + json.dumps(example["outputs"])
                for example in self.examples
            ]
        )
        return (
            f"Instructions:\n"
            f"{self.instructions}\n\n"
            f"Generate the response in JSON format with the following keys:\n"
            f"{json.dumps(self.outputs)}\n"
            f"Consider the instructions and the examples below to determine those values.\n\n"
            f"Examples:\n"
            f"{examples_section}\n\n"
            f"Inputs:\n"
            f"{inputs_section}\n"
            f"Outputs:\n"
        )

    @staticmethod
    def validate_input_parameters(
        expected: dict[str, Any], received: dict[str, Any]
    ) -> None:
        """
        Validate the input parameters.

        Args:
            expected (dict[str, Any]): The expected input parameters.
            received (dict[str, Any]): The received input parameters.

        Raises:
            ValueError: If not all expected inputs are present in the received inputs.
                        If the received inputs are not lists or have different lengths.
        """
        for param in expected.keys():
            if param not in received:
                msg = f"LLM evaluator expected input parameter '{param}' but received only {received.keys()}."
                raise ValueError(msg)

        if not all(isinstance(_input, list) for _input in received.values()):
            msg = (
                "LLM evaluator expects all input values to be lists but received "
                f"{[type(_input) for _input in received.values()]}."
            )
            raise ValueError(msg)

        inputs = received.values()
        length = len(next(iter(inputs)))
        if not all(len(_input) == length for _input in inputs):
            msg = (
                f"LLM evaluator expects all input lists to have the same length but received {inputs} with lengths "
                f"{[len(_input) for _input in inputs]}."
            )
            raise ValueError(msg)

    def is_valid_json_and_has_expected_keys(
        self, expected: list[str], received: str
    ) -> bool:
        """
        Output must be a valid JSON with the expected keys.

        Args:
            expected (list[str]): Names of expected outputs.
            received (str): Names of received outputs.

        Raises:
            ValueError: If the output is not a valid JSON with the expected keys:
                        - with `raise_on_failure` set to True a ValueError is raised.
                        - with `raise_on_failure` set to False a warning is issued and False is returned.

        Returns:
            bool: True if the received output is a valid JSON with the expected keys, False otherwise.
        """
        try:
            parsed_output = json.loads(received)
        except json.JSONDecodeError:
            msg = f"Response from LLM evaluator is not a valid JSON: {received}."
            if self.raise_on_failure:
                raise ValueError(msg)
            warn(msg)
            return False

        if not all(output in parsed_output for output in expected):
            msg = f"Expected response from LLM evaluator to be JSON with keys {expected}, got {received}."
            if self.raise_on_failure:
                raise ValueError(msg)
            warn(msg)
            return False

        return True

==> dynamiq-0.3.3.tar.gz: dynamiq-0.3.3/dynamiq/components/evaluators/llm_evaluator.py <==
import json
from typing import Any
from warnings import warn

from dynamiq.nodes import Node
from dynamiq.prompts import Message, Prompt


class LLMEvaluator:
    """
    Uses an LLM to evaluate inputs based on a prompt containing instructions and examples.

    The inputs are lists that are user-defined depending on the desired metric. The output
    is a dictionary with a key `results` containing a list of evaluation results. Each result
    is a dictionary with user-defined keys and values of either 0 for FALSE or 1 for TRUE.

    Usage example:
    ```python
    from typing import List
    from dynamiq.components.evaluators import LLMEvaluator
    evaluator = LLMEvaluator(
        instructions="Is this answer problematic for children?",
        inputs=[("predicted_answers", List[str])],
        outputs=["score"],
        examples=[
            {"inputs": {"predicted_answers": "Damn, this is straight outta hell!!!"},
             "outputs": {"score": 1}},
            {"inputs": {"predicted_answers": "Football is the most popular sport."},
             "outputs": {"score": 0}},
        ],
    )
    predicted_answers = [
        "Football is the most popular sport with around 4 billion followers worldwide",
        "Python language was created by Guido van Rossum.",
    ]
    results = evaluator.run(predicted_answers=predicted_answers)
    print(results)
    # {'results': [{'score': 0}, {'score': 0}]}
    ```
    """

    def __init__(
        self,
        instructions: str,
        inputs: list[tuple[str, type[list]]],
        outputs: list[str],
        examples: list[dict[str, Any]],
        *,
        raise_on_failure: bool = True,
        llm: Node,
    ):
        """
        Creates an instance of LLMEvaluator.

        Args:
            instructions (str): The prompt instructions to use for evaluation.
            inputs (list[tuple[str, type[list]]]): The inputs that the component expects as incoming
                connections and that it evaluates. Each input is a tuple of an input name and input
                type. Input types must be lists.
            outputs (list[str]): Output names of the evaluation results. They correspond to keys in
                the output dictionary.
            examples (list[dict[str, Any]]): Few-shot examples conforming to the expected input and
                output format as defined in the `inputs` and `outputs` parameters. Each example is a
                dictionary with keys "inputs" and "outputs". They contain the input and output as
                dictionaries respectively.
            raise_on_failure (bool): If True, the component will raise an exception on an
                unsuccessful API call.
            llm (Node): The LLM node to use for evaluation.
        """
        self.validate_init_parameters(inputs, outputs, examples)
        self.raise_on_failure = raise_on_failure
        self.instructions = instructions
        self.inputs = inputs
        self.outputs = outputs
        self.examples = examples
        self.api_params = {}

        default_generation_kwargs = {
            "response_format": {"type": "json_object"},
            "seed": 42,
        }
        user_generation_kwargs = self.api_params.get("generation_kwargs", {})
        merged_generation_kwargs = {
            **default_generation_kwargs,
            **user_generation_kwargs,
        }
        self.api_params["generation_kwargs"] = merged_generation_kwargs

        template = self.prepare_template()
        message = Message(role="user", content=template)
        self.prompt = Prompt(messages=[message])
        self.builder = Prompt(messages=[message])

        self.llm = llm

    @staticmethod
    def validate_init_parameters(
        inputs: list[tuple[str, type[list]]],
        outputs: list[str],
        examples: list[dict[str, Any]],
    ):
        """
        Validate the init parameters.

        Args:
            inputs (list[tuple[str, type[list]]]): The inputs to validate.
            outputs (list[str]): The outputs to validate.
            examples (list[dict[str, Any]]): The examples to validate.

        Raises:
            ValueError: If the inputs are not a list of tuples with a string and a type of list.
                        If the outputs are not a list of strings.
                        If the examples are not a list of dictionaries.
                        If any example does not have keys "inputs" and "outputs" with values that
                        are dictionaries with string keys.
        """
        if (
            not isinstance(inputs, list)
            or not all(isinstance(_input, tuple) for _input in inputs)
            or not all(
                isinstance(_input[0], str)
                and _input[1] is not list
                and len(_input) == 2
                for _input in inputs
            )
        ):
            msg = (
                f"LLM evaluator expects inputs to be a list of tuples. Each tuple must contain "
                f"an input name and type of list but received {inputs}."
            )
            raise ValueError(msg)

        if not isinstance(outputs, list) or not all(
            isinstance(output, str) for output in outputs
        ):
            msg = f"LLM evaluator expects outputs to be a list of str but received {outputs}."
            raise ValueError(msg)

        if not isinstance(examples, list) or not all(
            isinstance(example, dict) for example in examples
        ):
            msg = f"LLM evaluator expects examples to be a list of dictionaries but received {examples}."
            raise ValueError(msg)

        for example in examples:
            if (
                {"inputs", "outputs"} != example.keys()
                or not all(
                    isinstance(example[param], dict) for param in ["inputs", "outputs"]
                )
                or not all(
                    isinstance(key, str)
                    for param in ["inputs", "outputs"]
                    for key in example[param]
                )
            ):
                msg = (
                    f"LLM evaluator expects each example to have keys `inputs` and `outputs` with "
                    f"values that are dictionaries with str keys but received {example}."
                )
                raise ValueError(msg)

    def run(self, **inputs) -> dict[str, Any]:
        """
        Run the LLM evaluator.

        Args:
            inputs: The input values to evaluate. The keys are the input names and the values are
                lists of input values.

        Returns:
            dict[str, Any]: A dictionary with a `results` entry that contains a list of results.
                Each result is a dictionary containing the keys as defined in the `outputs` parameter
                of the LLMEvaluator and the evaluation results as the values. If an exception occurs
                for a particular input value, the result will be `None` for that entry. If the API is
                "openai" and the response contains a "meta" key, the metadata from OpenAI will be
                included in the output dictionary, under the key "meta".

        Raises:
            ValueError: Only in the case that `raise_on_failure` is set to True and the received
                inputs are not lists or have different lengths, or if the output is not a valid JSON
                or doesn't contain the expected keys.
        """
        self.validate_input_parameters(dict(self.inputs), inputs)

        input_names, values = inputs.keys(), list(zip(*inputs.values()))
        list_of_input_names_to_values = [dict(zip(input_names, v)) for v in values]

        results: list[dict[str, Any] | None] = []
        errors = 0
        for input_names_to_values in list_of_input_names_to_values:
            prompt = self.prompt
            try:
                result = self.llm.execute(
                    input_data=input_names_to_values, prompt=prompt
                )
            except Exception as e:
                msg = (
                    f"Error while generating response for prompt: {prompt}. Error: {e}"
                )
                if self.raise_on_failure:
                    raise ValueError(msg)
                warn(msg)
                results.append(None)
                errors += 1
                continue

            if self.is_valid_json_and_has_expected_keys(
                expected=self.outputs, received=result["content"]
            ):
                parsed_result = json.loads(result["content"])
                results.append(parsed_result)
            else:
                results.append(None)
                errors += 1

        if errors > 0:
            msg = f"LLM evaluator failed for {errors} out of {len(list_of_input_names_to_values)} inputs."
            warn(msg)

        return {"results": results}

    def prepare_template(self) -> str:
        """
        Prepare the prompt template.

        Combine instructions, inputs, outputs, and examples into one prompt template with the
        following format:
        Instructions:
        <instructions>

        Generate the response in JSON format with the following keys:
        <list of output keys>
        Consider the instructions and the examples below to determine those values.

        Examples:
        <examples>

        Inputs:
        <inputs>
        Outputs:

        Returns:
            str: The prompt template.
        """
        inputs_section = (
            "{"
            + ", ".join(
                [
                    f'"{input_socket[0]}": {{{{ {input_socket[0]} }}}}'
                    for input_socket in self.inputs
                ]
            )
            + "}"
        )

        examples_section = "\n".join(
            [
                "Inputs:\n"
                + json.dumps(example["inputs"])
                + "\nOutputs:\n"
                + json.dumps(example["outputs"])
                for example in self.examples
            ]
        )
        return (
            f"Instructions:\n"
            f"{self.instructions}\n\n"
            f"Generate the response in JSON format omit extra keys and markdown syntax elements, "
            f"and include the following keys:\n"
            f"{json.dumps(self.outputs)}\n"
            f"Consider the instructions and the examples below to determine those values.\n\n"
            f"Examples:\n"
            f"{examples_section}\n\n"
            f"Inputs:\n"
            f"{inputs_section}\n"
            f"Outputs:\n"
        )

    @staticmethod
    def validate_input_parameters(
        expected: dict[str, Any], received: dict[str, Any]
    ) -> None:
        """
        Validate the input parameters.

        Args:
            expected (dict[str, Any]): The expected input parameters.
            received (dict[str, Any]): The received input parameters.

        Raises:
            ValueError: If not all expected inputs are present in the received inputs.
                        If the received inputs are not lists or have different lengths.
        """
        for param in expected.keys():
            if param not in received:
                msg = f"LLM evaluator expected input parameter '{param}' but received only {received.keys()}."
                raise ValueError(msg)

        if not all(isinstance(_input, list) for _input in received.values()):
            msg = (
                "LLM evaluator expects all input values to be lists but received "
                f"{[type(_input) for _input in received.values()]}."
            )
            raise ValueError(msg)

        inputs = received.values()
        length = len(next(iter(inputs)))
        if not all(len(_input) == length for _input in inputs):
            msg = (
                f"LLM evaluator expects all input lists to have the same length but received {inputs} with lengths "
                f"{[len(_input) for _input in inputs]}."
            )
            raise ValueError(msg)

    def is_valid_json_and_has_expected_keys(
        self, expected: list[str], received: str
    ) -> bool:
        """
        Output must be a valid JSON with the expected keys.

        Args:
            expected (list[str]): Names of expected outputs.
            received (str): Names of received outputs.

        Raises:
            ValueError: If the output is not a valid JSON with the expected keys:
                        - with `raise_on_failure` set to True a ValueError is raised.
                        - with `raise_on_failure` set to False a warning is issued and False is returned.

        Returns:
            bool: True if the received output is a valid JSON with the expected keys, False otherwise.
        """
        try:
            parsed_output = json.loads(received)
        except json.JSONDecodeError:
            msg = f"Response from LLM evaluator is not a valid JSON: {received}."
            if self.raise_on_failure:
                raise ValueError(msg)
            warn(msg)
            return False

        if not all(output in parsed_output for output in expected):
            msg = f"Expected response from LLM evaluator to be JSON with keys {expected}, got {received}."
            if self.raise_on_failure:
                raise ValueError(msg)
            warn(msg)
            return False

        return True
